import rclpy
from rclpy.node import Node,SetParametersResult,Parameter
from sensor_msgs.msg import Image, CameraInfo
from cv_bridge import CvBridge
import threading
import time
from dataclasses import dataclass
from typing import List
import gc
from camera.cameras import (
    USB_Camera,
    CSI_Camera,
    Video_Input,
    RTSP_Stream,
    TOFCamera,
)
from camera.utils import (
    vertical_to_horizontal,
)
from lifecycle_msgs.msg import State
from lifecycle_msgs.msg import Transition
from camera.lifecycle import LifecycleNode
import cv2
from rclpy.qos import QoSPresetProfiles
DATA_QOS = QoSPresetProfiles.SENSOR_DATA.value

@dataclass
class Config:
    namespace: str
    rate: int
    type: str
    src: str
    height: int
    width: int

    
class CameraNode(LifecycleNode):
    def __init__(self):
        super().__init__('camera_node')
        self.declare_parameters(
            namespace='',
            parameters=[(attr, None) for attr in Config.__annotations__]
        )
        self.logger = self.get_logger()  
        self.config = None
        self.lock = threading.Lock()
        self.add_on_set_parameters_callback(self.config_update)

    def on_configure(self):                
        self.logger.info("on_configure() is called")
        self.config = Config(
            **{attr: self.get_parameter(attr).value for attr in Config.__annotations__}
        ) if self.config is None else self.config
        self.bridge = CvBridge()    
        try:
            with self.lock:
                self.camera = None
                self.restart_camera()
        except Exception as e:
            self.logger.error(f"Camera init failed: {e}")
        return Transition.TRANSITION_CALLBACK_SUCCESS
    
    def on_cleanup(self):
        self.logger.info("on_cleanup() is called")
        self.stop_camera()
        time.sleep(1)
        return Transition.TRANSITION_CALLBACK_SUCCESS    

    def on_activate(self):
        self.logger.info("on_activate() is called")
        self.camera_info_pub = self.create_publisher(CameraInfo, f'{self.config.namespace}/camera_info', QoSPresetProfiles.SERVICES_DEFAULT.value)
        self.image_pub = self.create_publisher(Image, f'{self.config.namespace}/image_raw', QoSPresetProfiles.SERVICES_DEFAULT.value)
        self.camera_info = CameraInfo()
        self.camera_info.height = self.config.height
        self.camera_info.width = self.config.width
        
        rate = self.config.rate
        self.timer = self.create_timer(1/rate, self.capture_frames)
        return Transition.TRANSITION_CALLBACK_SUCCESS
    
    def on_deactivate(self):
        self.logger.info("on_deactivate() is called")
        # Destroying any pub/sub/timer is buggy in ROS2
        # Don't destroy them, just set them to None and let the garbage collector do the job
        time.sleep(1)
        self.timer.cancel()
        # self.timer.destroy()
        self.timer = None
        
        # self.camera_info_pub.destroy()
        # self.image_pub.destroy()      
        self.camera_info_pub = None
        self.image_pub = None
        
        gc.collect() 
        return Transition.TRANSITION_CALLBACK_SUCCESS
    
    def on_shutdown(self):
        self.logger.info("on_shutdown() is called")
        return Transition.TRANSITION_CALLBACK_SUCCESS
    
    def destroy_node(self) -> bool:
        self.stop_camera()
        return super().destroy_node()
        
    def config_update(self, params: List[Parameter]):
        # if self.state != State.PRIMARY_STATE_ACTIVE:
        #     return SetParametersResult(successful=False, reason="Node is not active")

        self.logger.info(f'Update camera: {[str(p.name) for p in params]}')        
        for param in params:
            # apply the changes to the config
            if param.name in self.config.__annotations__:
                # If the value is the same, skip
                if param.value == self.config.__getattribute__(param.name):
                    continue
                self.config.__setattr__(param.name, param.value)
        self.logger.info(f"Camera Config: {self.config}")
        return SetParametersResult(successful=True)

    def restart_camera(self):
        self.stop_camera()
        if self.config.type == "usb":
            new_camera = USB_Camera()
        elif self.config.type == "csi":
            new_camera = CSI_Camera()
        elif self.config.type == "video":
            new_camera = Video_Input()
        elif self.config.type == "rtsp":
            new_camera = RTSP_Stream()
        elif self.config.type == "tof":
            new_camera = TOFCamera()    
        else:
            raise Exception(f"Unknown camera type: {self.config.type}")
        
        try:
            new_camera.open(self.config.src)
        except Exception as e:
            raise Exception(f"Unable to open camera: {e}")
        
        self.camera = new_camera
        self.logger.info("Camera Updated")
    
    def stop_camera(self):
        if hasattr(self, "camera") and self.camera is not None:
            self.camera.release()
            self.camera = None
            # Logic to stop the camera
            self.logger.info("Camera stopped")

    def capture_frames(self):
        if self.state != State.PRIMARY_STATE_ACTIVE or self.camera is None:
            return
        # loop the video stream
        if self.config.type == "video":
            if self.camera.video_capture.get(cv2.CAP_PROP_POS_FRAMES) == self.camera.video_capture.get(cv2.CAP_PROP_FRAME_COUNT):
                self.camera.video_capture.set(cv2.CAP_PROP_POS_FRAMES, 0)
        ret, frame = self.camera.read()            
        if not ret or frame is None:
            return
        frame = frame.get('rgb', None)
        frame = vertical_to_horizontal(frame)
        # frame=np.random.randint(0,255,(self.camera_info.height, self.camera_info.width, 3), dtype=np.uint8)
        frame = cv2.resize(frame, (self.camera_info.width, self.camera_info.height))
        
        # Convert the frame to ROS Image message
        image_msg = self.bridge.cv2_to_imgmsg(frame, encoding="bgr8")

        # Publish the image message
        self.image_pub.publish(image_msg)
        self.camera_info_pub.publish(self.camera_info)
    

def main(args=None):
    rclpy.init(args=args)
    node = CameraNode()
    try:
        rclpy.spin(node)
    except KeyboardInterrupt:        
        if node.state != State.PRIMARY_STATE_FINALIZED:
            node.stop_camera()
    rclpy.shutdown()
    
if __name__ == '__main__':
    main()
