
import cv2
import numpy as np
import json
try:
    import tof.tof_camera as tof_camera
except:
    pass
from camera.utils import INPUT_IMAGE_HEIGHT, INPUT_IMAGE_WIDTH

# from timeout_decorator import timeout, TimeoutError
import signal
class timeout:
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message
    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)
    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)
    def __exit__(self, type, value, traceback):
        signal.alarm(0)

class Camera:
    def __init__(self,):        
        # Initialize instance variables
        # OpenCV video capture element
        self.video_capture = None
        self.frame = np.zeros((INPUT_IMAGE_HEIGHT, INPUT_IMAGE_WIDTH,3),dtype=np.uint8)

    def get_video_portion(self):
        pass

    def read(self):
        try:
            grabbed, frame = self.video_capture.read()
            if not grabbed:
                frame = self.frame
            else:
                self.frame = frame
            frame_dict=dict(rgb=frame)
            return grabbed, frame_dict
        except:
            return False, None

    def release(self):
        if self.video_capture is not None:
            self.video_capture.release()
            self.video_capture = None

class USB_Camera(Camera):
    def open(self, src=0, frame_rate =30):
        # @timeout(3)
        # def opencv_cam(src):
        #     return cv2.VideoCapture(int(src))
        try:
            with timeout(3):
                self.video_capture =  cv2.VideoCapture(int(src))
            # Grab the first frame to start the video capturing
            self.grabbed, self.frame = self.video_capture.read()
        except Exception as e:
            self.video_capture = None
            raise e

class RTSP_Stream(Camera):
    def open(self, src=0):
        try:
            self.video_capture = cv2.VideoCapture(src)
            # Grab the first frame to start the video capturing
            self.grabbed, self.frame = self.video_capture.read()
        except RuntimeError:
            self.video_capture = None
            print("Unable to open RTSP_Stream")

class CSI_Camera(Camera):
    def open(self, *args, **kwargs):
        gstreamer_pipeline_string = self.gstreamer_pipeline(*args, **kwargs)
        try:
            self.video_capture = cv2.VideoCapture(
                gstreamer_pipeline_string, cv2.CAP_GSTREAMER
            )
            # Grab the first frame to start the video capturing
            self.grabbed, self.frame = self.video_capture.read()

        except RuntimeError:
            self.video_capture = None
    
    @staticmethod
    def gstreamer_pipeline(
        sensor_id=0,
        capture_width=1920,
        capture_height=1080,
        display_width=1920,
        display_height=1080,
        framerate=30,
        flip_method=0,
    ):
        return (f"""
            nvarguscamerasrc sensor-id={sensor_id} ! 
            video/x-raw(memory:NVMM), width=(int){capture_width}, height=(int){capture_height}, 
            framerate=(fraction){framerate}/1  ! 
            nvvidconv flip-method={flip_method} ! 
            video/x-raw, width=(int){display_width}, height=(int){display_height}, format=(string)BGRx ! 
            videoconvert ! 
            video/x-raw, format=(string)BGR ! appsink
            """
        )


class Video_Input(Camera):
    def open(self, src):
        try:
            self.video_capture = cv2.VideoCapture(src)
            # self.video_capture.set(1,500)
            # Grab the first frame to start the video capturing
            self.grabbed, self.frame = self.video_capture.read()

        except RuntimeError:
            self.video_capture = None
    
    def get_video_portion(self):
        portion = int(self.video_capture.get(cv2.CAP_PROP_POS_FRAMES))/int(self.video_capture.get(cv2.CAP_PROP_FRAME_COUNT))
        # simple progress bar
        bar_length = 20
        filled_length = int(portion * bar_length)
        bar = '█' * filled_length + '-' * (bar_length - filled_length)
        print("Video portion: ", end="")
        print(f'|{bar}| {portion:.0%} ', end='\r', flush=True)
        return portion

class TOFCamera(Camera):
    def __init__(self,):
        Camera.__init__(self)
        
        self.rgb = np.zeros((INPUT_IMAGE_HEIGHT, INPUT_IMAGE_WIDTH,3),dtype=np.uint8)
        self.rgbd =  np.zeros((INPUT_IMAGE_HEIGHT, INPUT_IMAGE_WIDTH,3),dtype=np.uint8)
        self.depth = np.zeros((INPUT_IMAGE_HEIGHT, INPUT_IMAGE_WIDTH,3),dtype=np.float32)
        self.gray = np.zeros((INPUT_IMAGE_HEIGHT, INPUT_IMAGE_WIDTH),dtype=np.uint8)
        
    def open(self):
        self.video_capture = tof_camera.CustomCamera()
        self.video_capture.start()

    def read(self):
        rgb = self.rgb
        rgbd = self.rgbd
        depth = self.depth
        gray = self.gray            
        frame_dict=dict(
            rgb=rgb,
            rgbd=rgbd,
            depth=depth,
            gray=gray
        )
        return True, frame_dict